/*
 * Copyright (c) 2023 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: Apache-2.0
 */
#ifndef H_EXAMPLE_MGMT_
#define H_EXAMPLE_MGMT_

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Group ID for example management group.
 */
#define MGMT_GROUP_ID_EXAMPLE  MGMT_GROUP_ID_PERUSER

/**
 * Command IDs for example management group.
 */
#define EXAMPLE_MGMT_ID_TEST   0
#define EXAMPLE_MGMT_ID_OTHER  1

/**
 * Command result codes for example management group.
 */
enum example_mgmt_err_code_t {
	/** No error, this is implied if there is no ret value in the response */
	EXAMPLE_MGMT_ERR_OK = 0,

	/** Unknown error occurred. */
	EXAMPLE_MGMT_ERR_UNKNOWN,

	/** The provided value is not wanted at this time. */
	EXAMPLE_MGMT_ERR_NOT_WANTED,

	/** The provided value was rejected by a hook. */
	EXAMPLE_MGMT_ERR_REJECTED_BY_HOOK,
};

typedef enum {
    BLE_DEV_ADDR_TYPE_PUBLIC = 0,
    BLE_DEV_ADDR_TYPE_RANDOM,
    BLE_DEV_ADDR_TYPE_NONE,
} ble_dev_addr_type_t;

#define BLE_MAX_ADV_REPORT_SIZE 31

typedef enum {
    BLE_ADV_TYPE_IND = 0,
    BLE_ADV_TYPE_DIRECT_IND,
    BLE_ADV_TYPE_SCAN_IND,
    BLE_ADV_TYPE_NONCONN_IND,
} ble_adv_type_t;

#pragma pack(1)

typedef struct {
    ble_dev_addr_type_t type;
    uint8_t mac_addr[6];
    uint8_t data[BLE_MAX_ADV_REPORT_SIZE];
    uint8_t data_len;
    int8_t rssi;
    ble_adv_type_t adv_type;
} ble_beacon_t;

#pragma pack()

void beacons_update(ble_beacon_t beacons_data);

#ifdef __cplusplus
}
#endif

#endif
