/*
 * Copyright (c) 2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

/** @file
 *  @brief Central Heart Rate over LE Coded PHY sample
 */

#include <zephyr/types.h>
#include <stddef.h>
#include <errno.h>
#include <zephyr/kernel.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/sys/printk.h>
#include <zephyr/sys/byteorder.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/conn.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/hci.h>
#include <bluetooth/gatt_dm.h>
#include <bluetooth/scan.h>
//#include <bluetooth/services/hrs_client.h>

#include <zephyr/usb/usb_device.h>
#include <zephyr/usb/usbd.h>

#include "example_mgmt.h"
#include "my_lbs.h"

const struct device *const rtc = DEVICE_DT_GET(DT_ALIAS(rtc));

#define BLE_INT_NODE DT_ALIAS(led0)
static const struct gpio_dt_spec led = GPIO_DT_SPEC_GET(BLE_INT_NODE, gpios);

static struct bt_conn *default_conn;

//static struct bt_hrs_client hrs_c;

static void app_led_cb(bool led_state)
{
	//dk_set_led(USER_LED, led_state);
	gpio_pin_set_dt(&led, led_state);
	printk("LED state changed: %s\n", led_state ? "ON" : "OFF");
}

static bool app_button_cb(void)
{
	printk("Button state request\n");
	return true;
}

static struct my_lbs_cb app_callbacks = {
	.led_cb    = app_led_cb,
	.button_cb = app_button_cb,
};

/*
static void notify_func(struct bt_hrs_client *hrs_c,
			const struct bt_hrs_client_measurement *meas,
			int err)
{
	if (err) {
		printk("Error during receiving Heart Rate Measurement notification, err: %d\n",
		       err);
		return;
	}

	printk("Heart Rate Measurement notification received:\n\n");
	printk("\tHeart Rate Measurement Value Format: %s\n",
		meas->flags.value_format ? "16 - bit" : "8 - bit");
	printk("\tSensor Contact detected: %d\n", meas->flags.sensor_contact_detected);
	printk("\tSensor Contact supported: %d\n", meas->flags.sensor_contact_supported);
	printk("\tEnergy Expended present: %d\n", meas->flags.energy_expended_present);
	printk("\tRR-Intervals present: %d\n", meas->flags.rr_intervals_present);

	printk("\n\tHeart Rate Measurement Value: %d bpm\n", meas->hr_value);

	if (meas->flags.energy_expended_present) {
		printk("\n\tEnergy Expended: %d J\n", meas->energy_expended);
	}

	if (meas->flags.rr_intervals_present) {
		printk("\t RR-intervals: ");
		for (size_t i = 0; i < meas->rr_intervals_count; i++) {
			printk("%d ", meas->rr_intervals[i]);
		}
	}

	printk("\n");
}
*/
/*
static void discover_hrs_completed(struct bt_gatt_dm *dm, void *ctx)
{
	int err;

	printk("The discovery procedure succeeded\n");

	bt_gatt_dm_data_print(dm);

	err = bt_hrs_client_handles_assign(dm, &hrs_c);
	if (err) {
		printk("Could not init HRS client object (err %d)\n", err);
		return;
	}

	err = bt_hrs_client_measurement_subscribe(&hrs_c, notify_func);
	if (err && err != -EALREADY) {
		printk("Subscribe failed (err %d)\n", err);
	} else {
		printk("[SUBSCRIBED]\n");
	}

	err = bt_gatt_dm_data_release(dm);
	if (err) {
		printk("Could not release the discovery data (err %d)\n", err);
	}
}

static void discover_hrs_service_not_found(struct bt_conn *conn, void *ctx)
{
	printk("No more services\n");
}

static void discover_hrs_error_found(struct bt_conn *conn, int err, void *ctx)
{
	printk("The discovery procedure failed, err %d\n", err);
}
*/
/*
static struct bt_gatt_dm_cb discover_hrs_cb = {
	.completed = discover_hrs_completed,
	.service_not_found = discover_hrs_service_not_found,
	.error_found = discover_hrs_error_found,
};*/

/*
static void scan_filter_match(struct bt_scan_device_info *device_info,
			      struct bt_scan_filter_match *filter_match,
			      bool connectable)
{
	//int err;
	char addr[BT_ADDR_LE_STR_LEN];
	//struct bt_conn_le_create_param *conn_params;
	static int number = 0;

	if (!connectable)
	{
		bt_addr_le_to_str(device_info->recv_info->addr, addr, sizeof(addr));

		printk("%d Filters matched. Address: %s connectable: %s\n",
			number++, addr, connectable ? "yes" : "no");

		ble_beacon_t beacons_data;

		beacons_data.type = BLE_DEV_ADDR_TYPE_PUBLIC;
		memcpy(beacons_data.mac_addr, device_info->recv_info->addr->a.val, 6);
		beacons_data.data_len = device_info->adv_data->len;
		memcpy(beacons_data.data, device_info->adv_data->data, device_info->adv_data->len);
		beacons_data.rssi = device_info->recv_info->rssi;
		beacons_data.adv_type = (ble_adv_type_t)device_info->recv_info->adv_type;
		//beacons_data.timestamp = k_uptime_get();

		beacons_update(beacons_data);
	}

#if 0
	err = bt_scan_stop();
	if (err) {
		printk("Stop LE scan failed (err %d)\n", err);
	}

	conn_params = BT_CONN_LE_CREATE_PARAM(
			BT_CONN_LE_OPT_CODED | BT_CONN_LE_OPT_NO_1M,
			BT_GAP_SCAN_FAST_INTERVAL,
			BT_GAP_SCAN_FAST_INTERVAL);

	err = bt_conn_le_create(device_info->recv_info->addr, conn_params,
				BT_LE_CONN_PARAM_DEFAULT,
				&default_conn);
	if (err) {
		printk("Create conn failed (err %d)\n", err);

		err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
		if (err) {
			printk("Scanning failed to start (err %d)\n", err);
			return;
		}
	}

	printk("Connection pending\n");
#endif
}

void filter_no_match(struct bt_scan_device_info *device_info, bool connectable)
{
	//int err;
	char addr[BT_ADDR_LE_STR_LEN];
	//struct bt_conn_le_create_param *conn_params;

	bt_addr_le_to_str(device_info->recv_info->addr, addr, sizeof(addr));

	printk("Filters no matched. Address: %s connectable: %s\n",
		addr, connectable ? "yes" : "no");
}
*/
//BT_SCAN_CB_INIT(scan_cb, scan_filter_match, NULL, NULL, NULL);
//BT_SCAN_CB_INIT(scan_cb, scan_filter_match, filter_no_match, NULL, NULL);

#if 1
static void scan_recv(const struct bt_le_scan_recv_info *info, struct net_buf_simple *buf)
{
	// if (info->adv_type & BT_GAP_ADV_TYPE_EXT_ADV &&
	//     info->adv_props & BT_GAP_ADV_PROP_EXT_ADV &&
	//     info->adv_props & BT_GAP_ADV_PROP_CONNECTABLE) {
	// 	/* Attempt connection request for device with extended advertisements */
	// 	memcpy(&ext_addr, info->addr, sizeof(ext_addr));
	// 	raise_evt(BT_SAMPLE_EVT_EXT_ADV_FOUND);
	// }
#if 0
	ble_beacon_t beacons_data;

	if (info->addr->type == BT_ADDR_LE_RANDOM)
	{
		//printk("Random address\n");
		beacons_data.type = BLE_DEV_ADDR_TYPE_RANDOM;
	}
	else
	{
		//printk("Public address\n");
		beacons_data.type = BLE_DEV_ADDR_TYPE_PUBLIC;
	}
	memcpy(beacons_data.mac_addr, info->addr->a.val, 6);
	beacons_data.data_len = buf->len;
	memcpy(beacons_data.data, buf->data, BLE_MAX_ADV_REPORT_SIZE > buf->len ? buf->len : BLE_MAX_ADV_REPORT_SIZE);
	beacons_data.rssi = info->rssi;
	beacons_data.adv_type = (ble_adv_type_t)info->adv_type;
	//beacons_data.timestamp = k_uptime_get();

	beacons_update(beacons_data);
#endif
	//printk("Adv receiving\n");
	//gpio_pin_toggle_dt(&led);
}
#endif

static struct bt_le_scan_cb scan_callbacks_recv = {
	.recv = scan_recv,
};
/*
static bt_addr_le_t local_addr = {
	//.a.val = { 0xC2, 0x03, 0x03, 0x00, 0x3A, 0x98 },
	.type = BT_ADDR_LE_RANDOM,
	//.a.val = { 0x98, 0x3A, 0x00, 0x03, 0x03, 0xC2 },
	.a = {
		.val = { 0x98, 0x3A, 0x00, 0x03, 0x03, 0xC2 },
	},
};*/ 

static void scan_init(void)
{
	int err;

	/* Use active scanning and disable duplicate filtering to handle any
	 * devices that might update their advertising data at runtime. */
	struct bt_le_scan_param scan_param = {
		//.type     = BT_LE_SCAN_TYPE_ACTIVE,
		.type     = BT_LE_SCAN_TYPE_PASSIVE,
		.options  = BT_LE_SCAN_OPT_CODED,// | BT_LE_SCAN_OPT_NO_1M,  // TODO: Enable back 1M PHY
		.interval = BT_GAP_SCAN_FAST_INTERVAL,
		.window   = BT_GAP_SCAN_FAST_WINDOW,
		//.interval = BT_GAP_SCAN_SLOW_INTERVAL_1,
		//.window   = BT_GAP_SCAN_SLOW_WINDOW_1,
	};

	struct bt_scan_init_param scan_init = {
		.scan_param = &scan_param,
		.connect_if_match = 0,
		.conn_param = NULL
	};

	bt_scan_init(&scan_init);
	//bt_scan_cb_register(&scan_cb);
	bt_le_scan_cb_register(&scan_callbacks_recv);

	//err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_ADDR, &local_addr);

	//err = bt_scan_filter_add(BT_SCAN_FILTER_TYPE_UUID, BT_UUID_HRS);
	// if (err) {
	// 	printk("Scanning filters cannot be set (err %d)\n", err);
	// 	return;
	// }

	// err = bt_scan_filter_enable(BT_SCAN_ADDR_FILTER, false);
	// if (err) {
	// 	printk("Filters cannot be turned on (err %d)\n", err);
	// }
}

static void connected(struct bt_conn *conn, uint8_t conn_err)
{
	int err;
	struct bt_conn_info info;
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	if (conn_err) {
		printk("Failed to connect to %s, 0x%02x %s\n", addr, conn_err,
		       bt_hci_err_to_str(conn_err));

		bt_conn_unref(default_conn);
		default_conn = NULL;

		err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
		if (err) {
			printk("Scanning failed to start (err %d)\n", err);
		}

		return;
	}

	err = bt_conn_get_info(conn, &info);
	if (err) {
		printk("Failed to get connection info (err %d)\n", err);
	} else {
		const struct bt_conn_le_phy_info *phy_info;

		phy_info = info.le.phy;
		printk("Connected: %s, tx_phy %u, rx_phy %u\n",
				addr, phy_info->tx_phy, phy_info->rx_phy);
	}

	/*
	if (conn == default_conn) {
		err = bt_gatt_dm_start(conn, BT_UUID_HRS, &discover_hrs_cb, NULL);
		if (err) {
			printk("Failed to start discovery (err %d)\n", err);
		}
	}*/
}

static void disconnected(struct bt_conn *conn, uint8_t reason)
{
	int err;
	char addr[BT_ADDR_LE_STR_LEN];

	bt_addr_le_to_str(bt_conn_get_dst(conn), addr, sizeof(addr));

	printk("Disconnected: %s, reason 0x%02x %s\n", addr, reason, bt_hci_err_to_str(reason));

	if (default_conn != conn) {
		return;
	}

	bt_conn_unref(default_conn);
	default_conn = NULL;

	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		printk("Scanning failed to start (err %d)\n", err);
	}
}

BT_CONN_CB_DEFINE(conn_callbacks) = {
	.connected = connected,
	.disconnected = disconnected,
};

static const struct bt_le_adv_param *adv_param = BT_LE_ADV_PARAM(
	(BT_LE_ADV_OPT_CONNECTABLE |
	 BT_LE_ADV_OPT_USE_IDENTITY), /* Connectable advertising and use identity address */
	800, /* Min Advertising Interval 500ms (800*0.625ms) */
	801, /* Max Advertising Interval 500.625ms (801*0.625ms) */
	NULL); /* Set to NULL for undirected advertising */

#define DEVICE_NAME CONFIG_BT_DEVICE_NAME
#define DEVICE_NAME_LEN (sizeof(DEVICE_NAME) - 1)

static const struct bt_data ad[] = {
	BT_DATA_BYTES(BT_DATA_FLAGS, (BT_LE_AD_GENERAL | BT_LE_AD_NO_BREDR)),
	BT_DATA(BT_DATA_NAME_COMPLETE, DEVICE_NAME, DEVICE_NAME_LEN),

};

static const struct bt_data sd[] = {
	BT_DATA_BYTES(BT_DATA_UUID128_ALL, BT_UUID_LBS_VAL),
};

int main(void)
{
	int err;

	err = usb_enable(NULL);
	if (err) {
		printk("USB init failed (err %d)\n", err);
		return 0;
	}

	if (!gpio_is_ready_dt(&led)) {
		printk("Device LED is not ready\n");
		return 0;
	}

	err = gpio_pin_configure_dt(&led, GPIO_OUTPUT_INACTIVE);
	if (err) {
		printk("Failed to configure LED pin (err %d)\n", err);
		return 0;
	}

	if (!device_is_ready(rtc)) {
		printk("Device RTC is not ready\n");
		return 0;
	}

	err = my_lbs_init(&app_callbacks);
	if (err) {
		printk("Failed to init LBS (err:%d)\n", err);
		return -1;
	}

	printk("Starting Bluetooth Central HR coded example\n");

	err = bt_enable(NULL);
	if (err) {
		printk("Bluetooth init failed (err %d)\n", err);
		return 0;
	}

	printk("Bluetooth initialized\n");

	// err = bt_hrs_client_init(&hrs_c);
	// if (err) {
	// 	printk("Heart Rate Service client failed to init (err %d)\n", err);
	// 	return 0;
	// }

	scan_init();

	err = bt_scan_start(BT_SCAN_TYPE_SCAN_ACTIVE);
	if (err) {
		printk("Scanning failed to start (err %d)\n", err);
		return 0;
	}

	printk("Scanning successfully started\n");

	err = bt_le_adv_start(adv_param, ad, ARRAY_SIZE(ad), sd, ARRAY_SIZE(sd));
	if (err) {
		printk("Advertising failed to start (err %d)\n", err);
		return 0;
	}

	printk("Advertising successfully started\n");

	while (1) {
		
		k_sleep(K_SECONDS(15));
		printk("System time: %d\n", k_uptime_get_32());
	}	

	return 0;
}
